package org.bitbucket.bradleysmithllc.etlunit.feature.informatica.util;

import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlunit.InformaticaError;
import org.bitbucket.bradleysmithllc.etlunit.Log;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.feature.informatica.InformaticaDomain;
import org.bitbucket.bradleysmithllc.etlunit.feature.informatica.InformaticaIntegrationService;
import org.bitbucket.bradleysmithllc.etlunit.feature.informatica.InformaticaRepository;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.Incomplete;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;
import org.bitbucket.bradleysmithllc.etlunit.util.regexp.*;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.*;

public final class InformaticaOutOfProcessRepositoryClient implements InformaticaRepositoryClient
{
	protected final Log applicationLog;

	public static final String FOLDER_LIST_ATTRIBUTE = "folderList";
	private final List<String> temporaryFolders = new ArrayList();

	private final InformaticaDomain informaticaDomain;

	private final InformaticaRepository informaticaRepository;
	private final RuntimeSupport runtimeSupport;

	private static int processNum = 0;

	private final int thisProcessNum = nextProcessNum();

	private static synchronized int nextProcessNum()
	{
		return processNum++;
	}

	private final PmRepProcess pmRep;

	public InformaticaOutOfProcessRepositoryClient(
		InformaticaRepository informaticaRepository,
		File pmCmdDir,
		RuntimeSupport pRuntimeSupport,
		Log log
	) throws Exception {
		applicationLog = log;
		this.informaticaRepository = informaticaRepository;
		this.informaticaDomain = informaticaRepository.getInformaticaDomain();
		runtimeSupport = pRuntimeSupport;

		Map<String, String> env = new HashMap<String, String>();

		env.put("INFA_DOMAINS_FILE", informaticaDomain.getDomainFile().getAbsolutePath());
		env.put("INFAPASSWD", informaticaDomain.getPasswordEncrypted());

		String pmRepId = informaticaDomain.getDomainName() + "." + informaticaRepository.getRepositoryName() + "." + thisProcessNum;
		File workingPmrep = new FileBuilder(runtimeSupport.getReportDirectory("log")).subdir("informatica").subdir("pmrepWorking." + pmRepId).mkdirs().file();

		// make sure the directory is empty
		FileUtils.cleanDirectory(workingPmrep);

		pmRep = new PmRepProcess(new File(pmCmdDir, "pmrep"), runtimeSupport, env, pmRepId, workingPmrep);

		connect();
	}

	public final void connect() throws Exception
	{
		String res = pmRep.send("connect -r " + informaticaRepository.getRepositoryName() + " -n " + informaticaDomain.getUserName() + " -X INFAPASSWD -d " + informaticaDomain.getDomainName());

		InfaConnectResultExpression icre = new InfaConnectResultExpression(res);

		if (!icre.hasNext())
		{
			throw new InformaticaError("Could not connect: " + res);
		}
	}

	@Incomplete
	public void createConnection(
		String serverName,
		String databaseName,
		String databaseUsername,
		String databasePassword,
		String relationalConnectionName
	) throws Exception, InformaticaError {
		String command = new StringBuilder("createconnection ")
			.append("-s ")
			.append("\"Microsoft SQL Server\" ")
			.append("-n ")
			.append(relationalConnectionName)
			.append(" -u ")
			.append(databaseUsername)
			.append(" -p ")
			.append(databasePassword)
			.append(" -v ")
			.append(serverName)
			.append(" -b ")
			.append(databaseName)
			.append(" -e")
			.append(" \"SET QUOTED_IDENTIFIER ON\"")
			.append(" -l")
			.append(" MS1252").toString();

		String outBuff = pmRep.send(command);

		InfaCreateConnectionResultExpression iccre = InfaCreateConnectionResultExpression.match(outBuff);

		if (!iccre.hasNext())
		{
			throw new InformaticaError("Could not create connection " + outBuff);
		}
	}

	public void createFolder(String folder) throws Exception, InformaticaError {
		createFolder(folder, false);
	}

	@Incomplete
	public void createFolder(String folder, boolean temporary) throws Exception, InformaticaError {
		String outBuff = pmRep.send("createfolder -n " + folder + " -s");

		// check for a successful result
		InfaCreateFolderResultExpression icfre = new InfaCreateFolderResultExpression(outBuff);

		if (!icfre.hasNext())
		{
			// check whether the folder already exists.  This is not an error.
			InfaCreateFolderExistsExpression icfee = InfaCreateFolderExistsExpression.match(outBuff, folder, informaticaRepository.getRepositoryName());

			if (!icfee.hasNext())
			{
				throw new InformaticaError("Could not create folder " + outBuff);
			}
		}

		if (temporary)
		{
			temporaryFolders.add(folder);
		}
	}

	@Incomplete
	public void deleteConnection(
		String connectionName
	) throws Exception, InformaticaError {
		deleteConnection(connectionName, true);
	}

	@Incomplete
	public void deleteConnection(
			String connectionName,
			boolean failIfNotExists
	) throws Exception, InformaticaError {
		String res = pmRep.send("deleteconnection -n " + connectionName + " -f");

		// look for success
		InfaDeleteConnectionSuccessExpression idcse = new InfaDeleteConnectionSuccessExpression(res);

		if (!idcse.hasNext())
		{
			// failed for some reason - look for a known reason (does not exist), and handle accordingly
			InfaDeleteConnectionFailedExpression idcfe = new InfaDeleteConnectionFailedExpression(res);

			if (idcfe.hasNext())
			{
				// look for the connection does not exist message
				InfaDeleteConnectionDoesNotExistExpression idcdnee = new InfaDeleteConnectionDoesNotExistExpression(res, connectionName);

				if (idcdnee.hasNext())
				{
					// handle according to the failIfNotExists parameter
					if (failIfNotExists)
					{
						throw new InformaticaError("Connection to delete does not exist");
					}
					else
					{
						applicationLog.info("User does not care if connection does not exist");
					}
				}
			}
		}
	}

	public void deleteFolder(String folder) throws Exception, InformaticaError {
		deleteFolder(folder, true);
	}

	@Incomplete
	public void deleteFolder(String folder, boolean failIfNotExists) throws Exception, InformaticaError {
		String res = pmRep.send("deletefolder -n " + folder);

		StringBuffer outBuff = new StringBuffer(res);

		InfaDeleteFolderSuccessExpression idfre = new InfaDeleteFolderSuccessExpression(outBuff);

		if (!idfre.hasNext())
		{
			// check for allowed not to exist
			InfaDeleteFolderFailedExpression idcfe = new InfaDeleteFolderFailedExpression(outBuff);

			if (idcfe.hasNext())
			{
				InfaDeleteFolderDoesNotExistExpression idcdnee = new InfaDeleteFolderDoesNotExistExpression(outBuff, folder);

				if (idcdnee.hasNext())
				{
					if (failIfNotExists)
					{
						throw new InformaticaError("Folder " + folder + " does not exist.");
					}
				}
				else
				{
					throw new InformaticaError("Could not delete folder " + outBuff);
				}
			}
			else
			{
				throw new InformaticaError("Could not delete folder " + outBuff);
			}
		}
	}

	@Incomplete
	public void deleteTemporaryFolders() throws Exception, InformaticaError {
		Iterator<String> it = temporaryFolders.iterator();

		while (it.hasNext())
		{
			String folder = it.next();

			deleteFolder(folder);

			it.remove();
		}
	}

	@Incomplete
	public void exportWorkflowToXml(
		String workFlow,
		String folder,
		File output
	) throws Exception, InformaticaError {
		String command = new StringBuilder("objectexport ")
			.append("-n ")
			.append("wkf_" + workFlow)
			.append(" -o ")
			.append("workflow ")
			.append("-m ")
			.append("-s ")
			.append("-r ")
			.append("-b ")
			.append("-f ")
			.append(folder)
			.append(" -u \"")
			.append(output.getAbsolutePath()).append("\"").toString();

		String outBuff = pmRep.send(command);

		InfaExportResultExpression iere = InfaExportResultExpression.match(outBuff.toString());

		if (!iere.hasNext())
		{
			throw new InformaticaError("Unexpected results from pmrep");
		}

		if ((iere.getErrors() + iere.getWarnings()) != 0)
		{
			throw new InformaticaError("Errors and/or warnings received from pmrep.");
		}

		File tempf = new File(output.getParentFile(), output.getName() + ".tmp");

		// correct silly stuff in xml output
		// generify repository name
		IOUtils.replace(output, tempf, informaticaRepository.getRepositoryName(), "svn_pc_repo");

		// generify integration service
		IOUtils.visitLines(tempf, output, new IOUtils.LineVisitor() {
			public String visit(String line, int lineNo) {
				InfaServerNameExpression isne = new InfaServerNameExpression(line);

				return isne.replaceAll("SERVERNAME =\"svn_is\"");
			}
		});

		// generify domain name
		IOUtils.visitLines(output, tempf, new IOUtils.LineVisitor() {
			public String visit(String line, int lineNo) {
				InfaDomainNameExpression isne = new InfaDomainNameExpression(line, informaticaDomain.getDomainName());

				return isne.replaceAll("SERVER_DOMAINNAME =\"Domain_etlsvn01\"");
			}
		});

		if (!output.delete())
		{
			throw new IOException("Obnoxious file won't delete!");
		}

		// strip out the time stamp
		IOUtils.replace(tempf, output, "<POWERMART CREATION_DATE=\"\\d{1,2}/\\d{1,2}/\\d{4} \\d{1,2}:\\d{1,2}:\\d{1,2}\" REPOSITORY_VERSION=\"(\\d{1,4}.\\d{1,9})\">", "<POWERMART CREATION_DATE=\"11/30/2010 12:00:00\" REPOSITORY_VERSION=\"$1\">");

		if (!tempf.delete())
		{
			throw new IOException("Obnoxious temp file won't delete!");
		}
	}

	@Incomplete
	public void exportAllWorkflowsToXml() throws Exception, InformaticaError {
		if (true) throw new UnsupportedOperationException();

		/*
		File queryf = runtimeSupport.createTempFile("ExportAll.txt");

		String command = new StringBuilder("executequery ")
			.append("-q ")
			.append("ExportAll ")
			.append("-t ")
			.append("shared ")
			.append("-u ")
			.append(queryf.getAbsolutePath()).toString();

		String output = pmRep.send(command);

		/**
		 * Handle this
		 * /
		//int res = process.getCompletionCode();

		//if (res != 0)
		//{
		//	throw new InformaticaError("Could not export from bitbucket " + process.getInputBuffered());
		//}

		// read the file line-by-line and export each workflow
		BufferedReader br = new BufferedReader(new FileReader(queryf));

		String line = null;
		int expCount = 0;

		while ((line = br.readLine()) != null)
		{
			expCount++;

			StringTokenizer st = new StringTokenizer(line, ",");

			String crap = st.nextToken();
			String folder = st.nextToken();
			String workflow = st.nextToken();

			// strip off the 'wkf_'
			if (!workflow.startsWith("wkf_"))
			{
				continue;
			}

			String workflowName = workflow.substring(4);

			String type = st.nextToken();
			String none = st.nextToken();
			String number = st.nextToken();

			exportWorkflowToXml(
				workflowName,
				folder,
				new File("")
			);
		}
		 */
	}

	@Incomplete
	public void importWorkflowFromXml(
		final String folderPrefix,
		InformaticaIntegrationService informaticaIntegrationService,
		File xmlFile,
		final List<String> flist
	) throws Exception {
		// prepare the import file
		File importFile = runtimeSupport.createTempFile(xmlFile.getName() + ".import");

		applicationLog.info("Substituting " + informaticaIntegrationService.getIntegrationServiceName() + " for svn_is");
		IOUtils.replace(xmlFile, importFile, "svn_is", informaticaIntegrationService.getIntegrationServiceName());

		// this needs to be a resource
		URL template = getClass().getResource("/objectimport-control.vm");

		if (template == null)
		{
			throw new IllegalStateException("Could not find objectimport-control.vm template");
		}

		File controlFile = runtimeSupport.createTempFile(xmlFile.getName() + ".control.xml");

		String insta = VelocityUtil.writeTemplate(IOUtils.readURLToString(template), new VelocityBean(folderPrefix, informaticaRepository.getRepositoryName(), flist));

		IOUtils.writeBufferToFile(controlFile, new StringBuffer(insta));

		Map<String, String> env = new HashMap<String, String>();

		String command = new StringBuilder("objectimport ")
			.append("-i \"")
			.append(importFile.getAbsolutePath())
			.append("\" -c \"")
			.append(controlFile.getAbsolutePath()).append("\"").toString();

		pmRep.send(command);

		/*
		int res = process.getCompletionCode();

		if (res != 0)
		{
			throw new InformaticaError("Could not import into bitbucket : objectimport returned " + res + " saying " + process.getInputBuffered());
		}
		 */

		importFile.delete();
	}

	public void importWorkflowFromXml(
		String folderPrefix,
		InformaticaIntegrationService informaticaIntegrationService,
		File xml
	) throws Exception {
		importWorkflowFromXml(folderPrefix, informaticaIntegrationService, xml, getFolderList(xml));
	}

	public static List<String> getFolderList(File xml) throws Exception {
		List<String> li = new ArrayList<String>();

		String str = IOUtils.readFileToString(xml);

		// list all folders
		FolderNameExpression fne = new FolderNameExpression(str);

		while (fne.hasNext()) {
			String name = fne.getFolderName();

			if (!li.contains(name)) {
				li.add(name);
			}
		}

		// list all folder shortcuts
		FolderShortcutExpression fse = new FolderShortcutExpression(str);

		while (fse.hasNext()) {
			String name = fse.getShortcutFolderName();

			if (!li.contains(name)) {
				li.add(name);
			}
		}

		return li;
	}

	public List<String> listFolders() throws Exception, InformaticaError {
		String res = pmRep.send("listobjects -o Folder");

		StringBuffer outBuff = new StringBuffer(res);
	
		List<String> folderList = new ArrayList<String>();

		InfaFolderNameExpression ifne = new InfaFolderNameExpression(outBuff.toString());

		while (ifne.hasNext())
		{
			String folder = ifne.getFolderName();

			folderList.add(folder);
		}

		return folderList;
	}

	public void purgeTestFolders() throws Exception {
		List<String> folderList = listFolders();

		Iterator<String> it = folderList.iterator();

		while (it.hasNext())
		{
			String folder = it.next();

			TestFolderNameExpression tfne = new TestFolderNameExpression(folder);

			if (tfne.matches())
			{
				deleteFolder(folder);
			}
		}
	}

	public InformaticaDomain getInformaticaDomain() {
		return informaticaDomain;
	}

	public InformaticaRepository getInformaticaRepository() {
		return informaticaRepository;
	}

	public void dispose()
	{
		try
		{
			pmRep.dispose();
		}
		catch (IOException e)
		{
			throw new IllegalStateException(e);
		}
	}
}